"""
The module contains PathFinder class which goes through dictionary of errors and gets all paths to elements.

"""

from copy import deepcopy
from functools import reduce
from operator import getitem
from typing import Any, Hashable, Union


class PathFinder:
    """
    PathFinder walks through error's dictionary and gets sequences of keys to all elements.

    """

    def __init__(self, errors: dict) -> None:
        """
        Initialize an object.

        Parameters
        ----------
        errors : Errors dictionary produced by Cerberus.

        """
        self._errors = errors
        self._paths = self._find_paths()

    @staticmethod
    def skip_lists(data: Union[dict, list]) -> Any:
        """
        Remove the most external lists from the object.

        Parameters
        ----------
        data : Input data, may be a dictionary or a list.

        Returns
        -------
        data : Output data without the most external lists.

        """
        while isinstance(data, list):
            if len(data) == 1:
                data = data[0]
            else:
                break

        return data

    def _getitem_skipping_lists(self, error: Union[dict, list], item: Hashable) -> Any:
        """
        Get item from a dictionary.

        """
        return getitem(self.skip_lists(error), item)

    def _get_element_by_path(self, error: Union[dict, list], path: tuple) -> Any:
        """
        Get an element by a path from nested dictionaries.

        """
        return self.skip_lists(
            reduce(self._getitem_skipping_lists, path, self.skip_lists(error))
        )

    def _del_element_by_path(self, error: Union[dict, list], path: tuple) -> None:
        """
        Remove an element by a path from nested dictionaries.

        """
        del self._get_element_by_path(error, path[:-1])[path[-1]]

    def _get_path_to_element(self, error: Union[dict, list]) -> tuple:
        """
        Get a path to an element in nested dictionaries.

        """
        path = list()
        error = self.skip_lists(error)

        while isinstance(error, dict):
            if error == {}:
                break

            key, error = error.popitem()
            error = self.skip_lists(error)
            path.append(key)

        return tuple(path)

    def _find_paths(self) -> tuple:
        """
        Find paths to all elements of error nested dictionaries.

        """
        paths = list()
        errors = deepcopy(self._errors)
        temp_errors = deepcopy(self._errors)

        while temp_errors != {}:
            path = self._get_path_to_element(temp_errors)

            if self._get_element_by_path(errors, path) != {}:
                paths.append(path)

            self._del_element_by_path(errors, path)
            temp_errors = deepcopy(errors)

        return tuple(paths)

    @property
    def paths(self) -> tuple:
        """
        Property of a list with all paths coming from an error dictionary.

        Returns
        -------
        tuple : A list of all paths.

        """
        return self._paths



"""
Unit tests for cerberror.trans module.

"""
from unittest.mock import MagicMock, Mock, patch, PropertyMock

import pytest

from cerberror.trans import Translator, ErrConverter
from tests.test_errors import path_to_file, ValidationError

_translate_result = {"_translate": "result"}
cerberus_errors_result = {"cerberus": "errors"}


class Validator:
    """Class emulating Validator of Cerberus."""

    def __init__(self, dct):
        self.document_error_tree = DocumentErrorTree(dct)


class DocumentErrorTree:
    """Class emulating Document Error Tree of Cerberus."""

    def __init__(self, dct):
        self._dct = self._convert(dct)

    @staticmethod
    def _convert(dct):
        for key in dct:
            dct.update({key: [ValidationError(i) for i in dct[key]]})

        return dct

    def fetch_errors_from(self, path):
        return self._dct[path]


@pytest.fixture
def path_finder_paths_mock():
    with patch(
        "cerberror.trans.PathFinder.paths", return_value=None, new_callable=PropertyMock
    ) as mock:
        yield mock


@pytest.fixture
def path_finder_init_mock(path_finder_paths_mock):
    with patch("cerberror.trans.PathFinder.__init__", return_value=None) as mock:
        yield mock


@pytest.fixture
def converter_records_mock():
    with patch(
        "cerberror.trans.ErrConverter.user_defined_records",
        return_value=None,
        new_callable=PropertyMock,
    ) as mock:
        yield mock


@pytest.fixture
def converter_init_mock(converter_records_mock):
    with patch("cerberror.trans.ErrConverter.__init__", return_value=None) as mock:
        yield mock


@pytest.fixture
def init_mock():
    with patch("cerberror.trans.Translator.__init__", return_value=None) as mock:
        yield mock


@pytest.fixture
def report_error_mock():
    with patch("cerberror.trans.Translator._report_error") as mock:
        yield mock


@pytest.fixture
def _translate_mock():
    with patch("cerberror.trans.Translator._translate") as mock:
        mock.return_value = _translate_result
        yield mock


@pytest.fixture
def translator_init_report_error_mock(
    init_mock,
    report_error_mock,
    path_finder_init_mock,
    path_finder_paths_mock,
    converter_init_mock,
    converter_records_mock,
):
    translator = Translator(Mock(), path_to_file)
    translator._path_to_file = path_to_file
    converter = ErrConverter(path_to_file)
    converter.any_error = False
    translator._converter = converter
    translator._validator = Mock()
    yield translator


@pytest.fixture
def translator_init_report_error__translate_mock(init_mock, report_error_mock, _translate_mock):
    converter, validator = MagicMock(), Mock()
    validator.errors = cerberus_errors_result

    translator = Translator(Mock(), path_to_file)
    translator._validator = validator
    translator._converter = converter
    yield translator


# ==================== TESTS ====================


def test_get_paths(translator_init_report_error_mock, report_error_mock, path_finder_paths_mock):
    translator_init_report_error_mock._get_paths()

    path_finder_paths_mock.assert_called_once()
    report_error_mock.assert_not_called()


def test_get_paths_fail(
    translator_init_report_error_mock, report_error_mock, path_finder_paths_mock
):
    path_finder_paths_mock.return_value = tuple()
    translator_init_report_error_mock._get_paths()

    path_finder_paths_mock.assert_called_once()
    report_error_mock.assert_called_once_with("No path was found")


def test_get_records(translator_init_report_error_mock, converter_records_mock):
    translator_init_report_error_mock._get_records()

    converter_records_mock.assert_called_once()


@pytest.mark.parametrize(
    "errors, result",
    [
        ("Error occurred", ["Error occurred"]),
        (["Error1", "Error2", "Error3"], ["Error1", "Error2", "Error3"]),
    ],
)
def test_report_error(errors, result):
    translator = Translator({}, path_to_file)

    if isinstance(errors, str):
        translator._report_error(errors)
    else:
        translator._report_error(*errors)

    assert translator._any_error
    assert translator._error_list == result


@pytest.mark.parametrize(
    "paths, pre_errors, records, sep, result",
    [
        (
            (("a", "b", "c", "d"), ("a", "b", "c", "e"), ("a", "b", "bb")),
            {
                ("a", "b", "c", "d"): [{"code": 66, "value": 3, "constraint": 5}],
                ("a", "b", "c", "e"): [{"code": 36}],
                ("a", "b", "bb"): [
                    {"code": 66, "value": 6, "constraint": 10},
                    {"code": 68, "value": 5, "constraint": [1, 2, 3]},
                ],
            },
            (
                (("a", "b", "c", "d"), 66, "{{value}} is less than {{constraint}}"),
                (("a", "b", "c", "e"), 36, "This number should be type of integer, bro!"),
                (("a", "b", "bb"), 66, "{{value}} cannot be less than {{constraint}}"),
                (("a", "b", "bb"), 68, "Only allowed values are {{constraint}}, not {{value}}"),
            ),
            " -> ",
            {
                "a -> b -> c -> d": ["3 is less than 5"],
                "a -> b -> c -> e": ["This number should be type of integer, bro!"],
                "a -> b -> bb": [
                    "6 cannot be less than 10",
                    "Only allowed values are [1, 2, 3], not 5",
                ],
            },
        ),
        (
            ((1, 2, "a"), ("b",), ("c", "d", "e")),
            {
                (1, 2, "a"): [{"code": 11, "value": 15.5, "type": "float"}],
                ("b",): [{"code": 2}, {"code": 404}],
                ("c", "d", "e"): [{"code": 28, "value": "Python", "constraint": "string"}],
            },
            (
                ((1, 2, "a"), 11, "{{value}} cannot be {{type}}"),
                (("b",), 2, "Awesome error"),
                (("c", "d", "e"), 28, "Value {{value}} is a {{constraint}}, it shouldn't be!"),
                (("b",), 404, "Another awesome error"),
            ),
            ".",
            {
                "1.2.a": ["15.5 cannot be float"],
                "b": ["Awesome error", "Another awesome error"],
                "c.d.e": ["Value Python is a string, it shouldn't be!"],
            },
        ),
    ],
)
def test__translate(
    translator_init_report_error_mock, report_error_mock, paths, pre_errors, records, sep, result
):
    translator_init_report_error_mock._validator = Validator(pre_errors)
    translator_init_report_error_mock._paths = paths
    translator_init_report_error_mock._records = records

    assert translator_init_report_error_mock._translate(sep) == result
    report_error_mock.assert_not_called()


@pytest.mark.parametrize(
    "paths, pre_errors, records, calls",
    [
        (
            (("a", "b", "c", "d"), ("a", "b", "c", "e"), ("a", "b", "bb")),
            {
                ("a", "b", "c", "d"): [{"code": 66, "value": 3, "constraint": 5}],
                ("a", "b", "c", "e"): [{"code": 36}],
                ("a", "b", "bb"): [
                    {"code": 66, "value": 6, "constraint": 10},
                    {"code": 68, "value": 5, "constraint": [1, 2, 3]},
                ],
            },
            (
                (("a", "b", "c", "d"), 66, "{{value}} is less than {{constraint}}"),
                (("a", "b", "bb"), 66, "{{value}} cannot be less than {{constraint}}"),
            ),
            2,
        ),
        (
            ((1, 2, "a"), ("b",), ("c", "d", "e")),
            {
                (1, 2, "a"): [{"code": 11, "value": 15.5, "type": "float"}],
                ("b",): [{"code": 2}, {"code": 404}],
                ("c", "d", "e"): [{"code": 28, "value": "Python", "constraint": "string"}],
            },
            (
                ((1, 2, "a"), 11, "{{value}} cannot be {{type}}"),
                (("b",), 2, "Awesome error"),
                (("c", "d", "e"), 28, "Value {{value}} is a {{constraint}}, it shouldn't be!"),
            ),
            1,
        ),
    ],
)
def test__translate_fail(
    translator_init_report_error_mock, report_error_mock, paths, pre_errors, records, calls
):
    translator_init_report_error_mock._validator = Validator(pre_errors)
    translator_init_report_error_mock._paths = paths
    translator_init_report_error_mock._records = records

    translator_init_report_error_mock._translate(">>")

    assert report_error_mock.call_count == calls


@pytest.mark.parametrize(
    "conv_any_err, trans_any_err, result",
    [
        (False, [False, False], _translate_result),
        (False, [False, True], cerberus_errors_result),
        (False, [True, True], cerberus_errors_result),
        (True, [False, False], cerberus_errors_result),
        (True, [True, True], cerberus_errors_result),
    ],
)
def test_translate_return(
    translator_init_report_error__translate_mock, conv_any_err, trans_any_err, result
):
    translator_init_report_error__translate_mock._converter.any_error = conv_any_err
    type(translator_init_report_error__translate_mock)._any_error = PropertyMock(
        side_effect=trans_any_err
    )

    assert translator_init_report_error__translate_mock.translate() == result


@pytest.mark.parametrize(
    "conv_any_err, trans_any_err", [(False, [False, False]), (False, [False, True])]
)
def test_translate(
    translator_init_report_error__translate_mock,
    report_error_mock,
    _translate_mock,
    conv_any_err,
    trans_any_err,
):
    translator_init_report_error__translate_mock._converter.any_error = conv_any_err
    type(translator_init_report_error__translate_mock)._any_error = PropertyMock(
        side_effect=trans_any_err
    )

    translator_init_report_error__translate_mock.translate()

    _translate_mock.assert_called_once()
    report_error_mock.assert_not_called()


@pytest.mark.parametrize(
    "conv_any_err, trans_any_err",
    [(False, [True, True]), (True, [False, False]), (True, [True, True])],
)
def test_translate_fail(
    translator_init_report_error__translate_mock,
    report_error_mock,
    _translate_mock,
    conv_any_err,
    trans_any_err,
):
    translator_init_report_error__translate_mock._converter.any_error = conv_any_err
    type(translator_init_report_error__translate_mock)._any_error = PropertyMock(
        side_effect=trans_any_err
    )

    translator_init_report_error__translate_mock.translate()

    _translate_mock.assert_not_called()
    report_error_mock.assert_called_once()



"""
Unit tests for cerberror.errors module.

"""
from io import StringIO
from unittest.mock import patch

import pytest

from cerberror.errors import ErrConverter

path_to_file = "path/to/file"


class ValidationError:
    """Class emulating ValidationError of Cerberus."""

    def __init__(self, attr_dct) -> None:
        for attr in attr_dct:
            setattr(self, attr, attr_dct[attr])


@pytest.fixture
def open_mock():
    with patch("cerberror.errors.open") as mock:
        yield mock


@pytest.fixture
def converter_init_mock():
    with patch("cerberror.errors.ErrConverter.__init__") as init_mock:
        init_mock.return_value = None
        converter = ErrConverter(path_to_file)
        converter._path_to_file = path_to_file
        converter._any_error = False
        yield converter


@pytest.fixture
def converter_report_error_mock(converter_init_mock):
    with patch("cerberror.errors.ErrConverter._report_error"):
        yield converter_init_mock


def test_read_predefined_messages(converter_init_mock, open_mock):
    stream = StringIO(
        "# path code message\n"
        "(1, 2, 'a')    67    \"My custom message for {{code}}\"\n"
        "# This is a comment\n"
        "     ('a', 'b', 'c') 2 \"Random message {{key}}\"\n"
        '(99,)   123 "Cannot be..."\n'
        "('key_a', 'key_b', 'key_c') 42 \n"
        "(7, 1, 'abc') 24 \"A new message (updated)\"\n"
        "('key_a', 'key_b', '7') 30  \"The {{last}} one message!\"\n"
    )
    result = (
        ((1, 2, "a"), 67, "My custom message for {{code}}"),
        (("a", "b", "c"), 2, "Random message {{key}}"),
        ((99,), 123, "Cannot be..."),
        ((7, 1, "abc"), 24, "A new message (updated)"),
        (("key_a", "key_b", "7"), 30, "The {{last}} one message!"),
    )
    open_mock.return_value = stream

    assert converter_init_mock._read_predefined_messages() == result


def test_read_predefined_messages_no_messages_found(converter_report_error_mock, open_mock):
    stream = StringIO(
        "# path code message\n"
        "!(1, 2, 'a')    67    \"My custom message for {{code}}\"\n"
        "# This is a comment\n"
        '$  (99,)   123 "Cannot be..."\n'
    )
    open_mock.return_value = stream
    converter_report_error_mock._read_predefined_messages()

    converter_report_error_mock._report_error.assert_called_once_with(
        f"No customized messages have been found in '{path_to_file}' file"
    )


def test_read_predefined_messages_file_not_found_error(converter_report_error_mock, open_mock):
    open_mock.side_effect = FileNotFoundError
    converter_report_error_mock._read_predefined_messages()

    converter_report_error_mock._report_error.assert_called_once_with(
        f"File '{path_to_file}' does not exist"
    )


@pytest.mark.parametrize(
    "error, predefined_msg, converted_msg",
    [
        (
            ValidationError({"field": "mass", "constraint": [100, 150, 200]}),
            "The {{field}} should be set on {{constraint}} only",
            "The mass should be set on [100, 150, 200] only",
        ),
        (
            ValidationError({"month": "February", "max": 29}),
            "{{month}} has max {{max}} days",
            "February has max 29 days",
        ),
        (
            ValidationError({"value": 53, "range": (12, 44)}),
            "{{value}} should be in range {{range}}",
            "53 should be in range (12, 44)",
        ),
        (ValidationError({}), "Example error message", "Example error message"),
    ],
)
def test_convert_message(converter_init_mock, error, predefined_msg, converted_msg):
    assert converter_init_mock.convert_message(error, predefined_msg) == converted_msg


@pytest.mark.parametrize(
    "error, predefined_msg, call_counter",
    [
        (
            ValidationError({"field": "mass", "constraint": [100, 150, 200]}),
            "The {{ field}} should be set on {{constraint}} only",
            1,
        ),
        (ValidationError({"month": "February", "max": 29}), "{{nonth}} has max {{ max }} days", 2),
    ],
)
def test_convert_message_no_attr(converter_report_error_mock, error, predefined_msg, call_counter):
    converter_report_error_mock.convert_message(error, predefined_msg)

    assert converter_report_error_mock._report_error.call_count == call_counter



"""
Cerberror
=========

The package converts errors generated by Cerberus
validator to customized messages defined by a user.

"""

__all__ = ["ErrConverter", "PathFinder", "Translator"]
__version__ = "0.1.1"
__author__ = "Przemysław Bruś"

from cerberror.errors import ErrConverter
from cerberror.paths import PathFinder
from cerberror.trans import Translator



"""
The module contains ErrConverter class replacing Cerberus errors to those defined by a user.

"""

import re
from ast import literal_eval
from pathlib import Path
from typing import Union

from cerberus.errors import ValidationError


class ErrConverter:
    """
    ErrConverter converts errors produced by Cerberus to customized messages.

    """

    def __init__(self, path_to_file: Union[str, Path]) -> None:
        """
        Initialize an object.

        Parameters
        ----------
        path_to_file : A name of the file storing customized error messages.

        """
        self._path_to_file = Path(path_to_file)
        self.any_error = False
        self.error_list = list()
        self._user_defined_records = self._read_predefined_messages()

    def _read_predefined_messages(self) -> tuple:
        """
        Read records from a file containing customized errors.

        """
        records = list()

        try:
            with open(self._path_to_file, "r") as file:
                for line in file:
                    record = re.split(
                        r"^\s*([(].*[,].*[)])\s+(\d+)\s+([\"].+[\"])", line.strip()
                    )[1:-1]
                    if record != list():
                        records.append(tuple([i for i in map(literal_eval, record)]))

            if len(records) == 0:
                self._report_error(
                    f"No customized messages have been found in '{self._path_to_file}' file"
                )

        except FileNotFoundError:
            self._report_error(f"File '{self._path_to_file}' does not exist")

        return tuple(records)

    def _report_error(self, error: str) -> None:
        """
        Notify occurred errors.

        """
        self.any_error = True
        self.error_list.append(error)

    def convert_message(self, error: ValidationError, message: str) -> str:
        """
        Convert a predefined message. This method replaces expressions within double curly brackets of a predefined
        message with counterparts from ValidationError class.

        Parameters
        ----------
        error : ValidationError object from Cerberus.
        message : Predefined message defined by a user.

        Returns
        -------
        message : Error message defined by a user. None if an error occurs.

        """
        any_error = False
        attributes = [i.strip("{}") for i in re.findall(r"{{[^{}]+}}", message)]

        for attr in attributes:
            if hasattr(error, attr):
                message = message.replace("{{" + attr + "}}", str(getattr(error, attr)))
            else:
                any_error = True
                self._report_error(
                    f"Invalid expression '{{{{{attr}}}}}' in file '{self._path_to_file}'"
                )

        return message if not any_error else None

    @property
    def user_defined_records(self) -> tuple:
        """
        A property for user defined records.

        Returns
        -------
        tuple : A list of records. Each record consists of:
                - path to message
                - code of error
                - predefined message

        """
        return self._user_defined_records



"""
This module contains Translator class which translates errors generated by Cerberus to messages defined by a user.

"""

from pathlib import Path
from typing import Union

from cerberus import Validator

from cerberror.errors import ErrConverter
from cerberror.paths import PathFinder


class Translator:
    """
    Translator allows to customize error messages produced by the Cerberus Validator.

    """

    def __init__(self, validator: Validator, path_to_file: Union[str, Path]) -> None:
        """
        Initialize an object and trigger internal computations.

        Parameters
        ----------
        validator : Cerberus object.
        path_to_file : A name of the file storing customized error messages.

        """
        self._validator = validator
        self._path_to_file = Path(path_to_file)
        self._converter = ErrConverter(self._path_to_file)
        self._any_error = False
        self._error_list = list()
        self._paths = None
        self._records = None
        self._errors = dict()

    def _get_paths(self) -> tuple:
        """
        Get paths to all errors produced by Cerberus.

        """
        path_finder = PathFinder(self._validator.errors)
        self._paths = path_finder.paths

        if self._paths == ():
            self._report_error("No path was found")

        return self._paths

    def _get_records(self) -> tuple:
        """
        Get records (paths, err codes, messages) defined by user.

        """
        self._records = self._converter.user_defined_records

        return self._records

    def translate(self, sep: str = " -> ") -> dict:
        """
        Translate errors generated by Cerberus into messages defined by a user.

        Parameters
        ----------
        sep : A string separator between elements in paths. The default is " -> ".

        Returns
        -------
        errors : If success, a result is a dictionary composed of pairs (path to element):(list of errors).
                 Otherwise the returned value is an error container generated by Cerberus originally.

        """
        if self._converter.any_error or self._any_error:
            self._report_error(*self._converter.error_list)
            self._errors = self._validator.errors
        else:
            self._errors = self._translate(sep)

        if self._any_error:
            self._errors = self._validator.errors

        return self._errors

    def _translate(self, sep: str) -> dict:
        """
        Translate errors into defined messages.

        """
        errors = dict()

        for path in self.paths:
            for error in self._validator.document_error_tree.fetch_errors_from(path):
                no_record = False
                for record in self.records:
                    if (error.code in record) and (path in record):
                        no_record = True
                        key = sep.join(map(str, path))

                        if key not in errors:
                            errors.update({key: [self._converter.convert_message(error, record[-1])]})
                        else:
                            errors[key].append(self._converter.convert_message(error, record[-1]))

                if not no_record:
                    self._report_error(
                        f"File '{self._path_to_file}' does not contain a record "
                        f"for path {path} and error code {error.code}"
                    )

        if self._converter.any_error:
            self._report_error(*self._converter.error_list)

        return errors

    def _report_error(self, *errors: str) -> None:
        """
        Notify occurred errors.

        """
        self._any_error = True

        for error in errors:
            self._error_list.append(error)

    @property
    def paths(self) -> tuple:
        """
        Get paths to all elements of Cerberus error contained in nested dictionaries.

        Returns
        -------
        tuple : A list of all paths.

        """
        if (self._paths is None) and (not self._any_error):
            self._get_paths()

        return self._paths

    @property
    def records(self) -> tuple:
        """
        Get records defined by a user in a text file.

        Returns
        -------
        tuple : A list of records. Each record consists of:
                - path to message
                - code of error
                - predefined message

        """
        if (self._records is None) and (not self._any_error):
            self._get_records()

        return self._records

    @property
    def errors(self) -> dict:
        """
        Get converted errors generated by Cerberus.

        Returns
        -------
        dict : A dictionary composed of pairs (path to element):(list of errors) or errors returned by Cerberus.

        """
        return self._errors

    @property
    def any_error(self) -> bool:
        """
        Check whether any error occurred while converting Cerberus's errors.

        Returns
        -------
        bool : False if no errors, otherwise True.

        """
        return self._any_error

    @property
    def error_list(self) -> list:
        """
        Get list of errors which occurred during converting Cerberus's errors into user-defined messages.

        Returns
        -------
        list : List of messages.

        """
        return self._error_list

    @property
    def validator(self) -> Validator:
        """
        Get validator of Cerberus.

        Returns
        -------
        Validator : Cerberus's validator.

        """
        return self._validator

    @validator.setter
    def validator(self, new_validator: Validator) -> None:
        """
        Setter for validator.

        """
        self.__init__(new_validator, self._path_to_file)

    @property
    def path_to_file(self) -> Path:
        """
        Get path to file which stores user defined records.

        Returns
        -------
        Path : path to the file with customized messages.

        """
        return self._path_to_file

    @path_to_file.setter
    def path_to_file(self, new_path_to_file) -> None:
        """
        Setter for path_to_file.

        """
        self.__init__(self._validator, new_path_to_file)
